@@ -420,7 +420,7 @@ static int probe_all(blkid_cache cache, int only_if_new)
 {
 	FILE *proc;
 	char line[1024];
-	char ptname0[128], ptname1[128], *ptname = 0;
+	char ptname0[128 + 1], ptname1[128 + 1], *ptname = 0;
 	char *ptnames[2];
 	dev_t devs[2];
 	int ma, mi;
@@ -550,7 +550,7 @@ static int probe_all_removable(blkid_cache cache)
 		return -BLKID_ERR_PROC;
 
 	while((d = readdir(dir))) {
-		struct sysfs_cxt sysfs;
+		struct sysfs_cxt sysfs = UL_SYSFSCXT_EMPTY;
 		int removable = 0;
 		dev_t devno;
 
