@@ -41,6 +41,7 @@
 #include "usleep.h"
 #include "strutils.h"
 #include "c.h"
+#include "closestream.h"
 
 /* constants from legacy PC/AT hardware */
 #define	RTC_PF	0x40
@@ -49,7 +50,6 @@
 
 #define MAX_LINE		1024
 
-#define VERSION_STRING		"rtcwake from " PACKAGE_STRING
 #define RTC_PATH		"/sys/class/rtc/%s/device/power/wakeup"
 #define SYS_POWER_STATE_PATH	"/sys/power/state"
 #define ADJTIME_PATH		"/etc/adjtime"
@@ -83,11 +83,11 @@ static struct option long_options[] = {
 
 static void __attribute__((__noreturn__)) usage(FILE *out)
 {
-	fputs(_("\nUsage:\n"), out);
+	fputs(USAGE_HEADER, out);
 	fprintf(out,
 	      _(" %s [options]\n"), program_invocation_short_name);
 
-	fputs(_("\nOptions:\n"), out);
+	fputs(USAGE_OPTIONS, out);
 	fputs(_(" -d, --device <device>    select rtc device (rtc0|rtc1|...)\n"
 		" -n, --dry-run            does everything, but suspend\n"
 		" -l, --local              RTC uses local timezone\n"
@@ -95,10 +95,13 @@ static void __attribute__((__noreturn__)) usage(FILE *out)
 		" -s, --seconds <seconds>  seconds to sleep\n"
 		" -t, --time <time_t>      time to wake\n"
 		" -u, --utc                RTC uses UTC\n"
-		" -v, --verbose            verbose messages\n"
-		" -V, --version            show version\n"), out);
+		" -v, --verbose            verbose messages\n"), out);
 
-	fputs(_("\nFor more information see rtcwake(8).\n"), out);
+	printf(USAGE_SEPARATOR);
+	printf(USAGE_HELP);
+	printf(USAGE_VERSION);
+
+	printf(USAGE_MAN_TAIL("rtcwake(8)"));
 
 	exit(out == stderr ? EXIT_FAILURE : EXIT_SUCCESS);
 }
@@ -112,7 +115,7 @@ static int is_wakeup_enabled(const char *devname)
 	snprintf(buf, sizeof buf, RTC_PATH, devname + strlen("/dev/"));
 	f = fopen(buf, "r");
 	if (!f) {
-		warn(_("open failed: %s"), buf);
+		warn(_("cannot open %s"), buf);
 		return 0;
 	}
 	s = fgets(buf, sizeof buf, f);
@@ -271,7 +274,7 @@ static void suspend_system(const char *suspend)
 	FILE	*f = fopen(SYS_POWER_STATE_PATH, "w");
 
 	if (!f) {
-		warn(_("open failed: %s"), SYS_POWER_STATE_PATH);
+		warn(_("cannot open %s"), SYS_POWER_STATE_PATH);
 		return;
 	}
 
@@ -281,7 +284,8 @@ static void suspend_system(const char *suspend)
 	}
 
 	/* this executes after wake from suspend */
-	fclose(f);
+	if (close_stream(f))
+		errx(EXIT_FAILURE, _("write error"));
 }
 
 
@@ -390,6 +394,7 @@ int main(int argc, char **argv)
 	setlocale(LC_ALL, "");
 	bindtextdomain(PACKAGE, LOCALEDIR);
 	textdomain(PACKAGE);
+	atexit(close_stdout);
 
 	while ((t = getopt_long(argc, argv, "ahd:lm:ns:t:uVv",
 					long_options, NULL)) != EOF) {
@@ -438,16 +443,14 @@ int main(int argc, char **argv)
 
 			/* alarm time, seconds-to-sleep (relative) */
 		case 's':
-			seconds = strtol_or_err(optarg,
-					_("failed to parse seconds value"));
+			seconds = strtou32_or_err(optarg, _("invalid seconds argument"));
 			break;
 
 			/* alarm time, time_t (absolute, seconds since
 			 * 1/1 1970 UTC)
 			 */
 		case 't':
-			alarm = strtol_or_err(optarg,
-					_("failed to parse time_t value"));
+			alarm = strtou32_or_err(optarg, _("invalid time argument"));
 			break;
 
 		case 'u':
@@ -459,7 +462,7 @@ int main(int argc, char **argv)
 			break;
 
 		case 'V':
-			printf("%s\n", VERSION_STRING);
+			printf(UTIL_LINUX_VERSION);
 			exit(EXIT_SUCCESS);
 
 		case 'h':
@@ -495,7 +498,6 @@ int main(int argc, char **argv)
 
 		strcpy(new_devname, "/dev/");
 		strcat(new_devname, devname);
-		free(devname);
 		devname = new_devname;
 	}
 
@@ -510,7 +512,7 @@ int main(int argc, char **argv)
 	fd = open(devname, O_RDONLY);
 #endif
 	if (fd < 0)
-		err(EXIT_FAILURE, _("open failed: %s"), devname);
+		err(EXIT_FAILURE, _("cannot open %s"), devname);
 
 	/* relative or absolute alarm time, normalized to time_t */
 	if (get_basetimes(fd) < 0)
@@ -612,8 +614,25 @@ int main(int argc, char **argv)
 		suspend_system(suspend);
 	}
 
-	if (!dryrun && ioctl(fd, RTC_AIE_OFF, 0) < 0)
-		warn(_("disable rtc alarm interrupt failed"));
+	if (!dryrun) {
+		/* try to disable the alarm with the preferred RTC_WKALM_RD and
+		 * RTC_WKALM_SET calls, if it fails fall back to RTC_AIE_OFF
+		 */
+		struct rtc_wkalrm wake;
+
+		if (ioctl(fd, RTC_WKALM_RD, &wake) < 0) {
+			if (ioctl(fd, RTC_AIE_OFF, 0) < 0) {
+				warn(_("disable rtc alarm interrupt failed"));
+				rc = EXIT_FAILURE;
+			}
+		} else {
+			wake.enabled = 0;
+			if (ioctl(fd, RTC_WKALM_SET, &wake) < 0) {
+				warn(_("disable rtc alarm interrupt failed"));
+				rc = EXIT_FAILURE;
+			}
+		}
+	}
 
 	close(fd);
 	return rc;
